"""Utility to create an argument parser"""

import argparse

from metrontagger import __version__


def make_parser() -> argparse.ArgumentParser:
    """Function to create the argument parser"""
    parser = argparse.ArgumentParser(
        description="Read in a file or set of files, and return the result.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument("path", nargs="+", help="Path of a file or a folder of files.")
    parser.add_argument(
        "-r",
        "--rename",
        help="Rename comic archive from the files metadata.",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "-o",
        "--online",
        help="Search online and attempt to identify comic archive.",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "-m",
        "--metroninfo",
        help="Write, delete, or validate MetronInfo.xml.",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "-c",
        "--comicinfo",
        help="Write, delete, or validate ComicInfo.xml.",
        action="store_true",
        default=False,
    )
    parser.add_argument("--id", help="Identify file for tagging with the Metron Issue Id.")
    parser.add_argument(
        "-d",
        "--delete",
        help="Delete the metadata tags from the file.",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "--ignore-existing",
        help="Ignore files that have existing metadata tag.",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "-i",
        "--interactive",
        help="Interactively query the user when there are matches for an online search.",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "--missing",
        help="List files without metadata.",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "-s",
        "--sort",
        help="Sort files that contain metadata tags.",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "-z",
        "--export-to-cbz",
        help="Export a CBR (rar) archive to a CBZ (zip) archive.",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "--validate",
        help="Verify that comic archive has a valid metadata xml.",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "--remove-non-valid",
        help="Remove metadata xml from comic if not valid. Used with --validate option",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "--delete-original",
        help="Delete the original archive after successful export to another format.",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "--duplicates",
        help="Identify and give the option to delete duplicate pages "
        "in a directory of comics. (Experimental)",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "--version",
        action="version",
        version=f"%(prog)s {__version__}",
        help="Show the version number and exit",
    )

    return parser



from __future__ import annotations

import sys
from logging import getLogger
from pathlib import Path
from typing import TYPE_CHECKING

import questionary
from darkseid.comic import Comic, MetadataFormat
from darkseid.metadata import Metadata
from darkseid.utils import get_recursive_filelist
from tqdm import tqdm

from metrontagger.duplicates import DuplicateIssue, Duplicates
from metrontagger.filerenamer import FileRenamer
from metrontagger.filesorter import FileSorter
from metrontagger.logging import init_logging

if TYPE_CHECKING:
    from metrontagger.settings import MetronTaggerSettings
from metrontagger.styles import Styles
from metrontagger.talker import Talker
from metrontagger.validate import SchemaVersion, ValidateMetadata

LOGGER = getLogger(__name__)


class Runner:
    """Class for running Metron Tagger operations.

    This class handles the execution of various operations such as renaming comics, exporting to cbz,
    removing duplicates, validating ComicInfo, and more based on the provided settings.
    """

    def __init__(self: Runner, config: MetronTaggerSettings) -> None:
        """Initialize the Runner with MetronTaggerSettings.

        This method sets up the Runner object with the provided MetronTaggerSettings configuration.
        """

        self.config = config

    def rename_comics(self: Runner, file_list: list[Path]) -> list[Path]:
        """Rename comic archives based on metadata.

        This method renames comic archives in the provided file list according to the metadata information,
        using the specified renaming template and settings.

        Args:
            file_list: list[Path]: The list of comic archive file paths to rename.

        Returns:
            list[Path]: The updated list of file paths after renaming.
        """

        questionary.print(
            "\nStarting comic archive renaming:\n-------------------------------",
            style=Styles.TITLE,
        )

        # Lists to track filename changes
        new_file_names: list[Path] = []
        original_files_changed: list[Path] = []
        renamer = FileRenamer()
        for comic in file_list:
            comic_archive = Comic(comic)
            if comic_archive.has_metadata(MetadataFormat.METRON_INFO):
                md = comic_archive.read_metadata(MetadataFormat.METRON_INFO)
            elif comic_archive.has_metadata(MetadataFormat.COMIC_RACK):
                md = comic_archive.read_metadata(MetadataFormat.COMIC_RACK)
            else:
                questionary.print(
                    f"skipping '{comic.name}'. no metadata available.",
                    style=Styles.WARNING,
                )
                continue

            renamer.set_metadata(md)
            renamer.set_template(self.config.rename_template)
            renamer.set_issue_zero_padding(self.config.rename_issue_number_padding)
            renamer.set_smart_cleanup(self.config.rename_use_smart_string_cleanup)

            unique_name = renamer.rename_file(comic)
            if unique_name is None:
                continue

            # track what files are being renamed
            new_file_names.append(unique_name)
            original_files_changed.append(comic)

            questionary.print(
                f"renamed '{comic.name}' -> '{unique_name.name}'",
                style=Styles.SUCCESS,
            )

        # Update file_list for renamed files
        for original_file in original_files_changed:
            file_list.remove(original_file)

        # Add new file names to file list.
        file_list.extend(iter(new_file_names))
        return file_list

    def _export_to_zip(self: Runner, file_list: list[Path]) -> None:
        """Export comic archives to cbz format.

        This method exports the comic archives in the provided file list to cbz format, and optionally deletes the
        original archives based on the configuration settings.

        Args:
            file_list: list[Path]: The list of comic archive file paths to export.

        Returns:
            None
        """

        questionary.print("\nExporting to cbz:\n-----------------", style=Styles.TITLE)
        for comic in file_list:
            ca = Comic(str(comic))
            if ca.is_rar():
                new_fn = Path(comic).with_suffix(".cbz")
                if ca.export_as_zip(new_fn):
                    questionary.print(
                        f"Exported '{comic.name}' to a cbz archive.",
                        style=Styles.SUCCESS,
                    )
                    if self.config.delete_original:
                        questionary.print(f"Removing '{comic.name}'.", style=Styles.SUCCESS)
                        comic.unlink()
                else:
                    questionary.print(f"Failed to export '{comic.name}'", style=Styles.ERROR)
            else:
                questionary.print(
                    f"'{comic.name}' is not a cbr archive. skipping...",
                    style=Styles.WARNING,
                )

    def _validate_comic_info(
        self: Runner, file_list: list[Path], remove_ci: bool = False
    ) -> None:
        """Validate ComicInfo metadata in comic archives."""

        questionary.print("\nValidating ComicInfo:\n---------------------", style=Styles.TITLE)
        for comic in file_list:
            ca = Comic(comic)
            has_comic_rack = ca.has_metadata(MetadataFormat.COMIC_RACK)
            has_metron_info = ca.has_metadata(MetadataFormat.METRON_INFO)

            if not has_comic_rack and not has_metron_info:
                questionary.print(
                    f"'{ca.path.name}' doesn't have any metadata files.",
                    style=Styles.WARNING,
                )
                continue

            if self.config.use_comic_info and has_comic_rack:
                xml = ca.archiver.read_file("ComicInfo.xml")
                self._check_if_xml_is_valid(ca, xml, MetadataFormat.COMIC_RACK, remove_ci)

            if self.config.use_metron_info and has_metron_info:
                xml = ca.archiver.read_file("MetronInfo.xml")
                self._check_if_xml_is_valid(ca, xml, MetadataFormat.METRON_INFO, remove_ci)

    @staticmethod
    def _check_if_xml_is_valid(
        comic: Comic, xml: bytes, fmt: MetadataFormat, remove_metadata: bool
    ) -> None:
        result = ValidateMetadata(xml).validate()
        messages = {
            SchemaVersion.ci_v2: (
                f"'{comic.path.name}' has a valid ComicInfo Version 2",
                Styles.SUCCESS,
            ),
            SchemaVersion.ci_v1: (
                f"'{comic.path.name}' has a valid ComicInfo Version 1",
                Styles.SUCCESS,
            ),
            SchemaVersion.mi_v1: (
                f"'{comic.path.name}' has a valid MetronInfo Version 1",
                Styles.SUCCESS,
            ),
        }

        message, style = messages.get(
            result, (f"'{comic.path.name}' is not valid", Styles.ERROR)
        )
        questionary.print(message, style=style)

        if result not in messages and remove_metadata and comic.remove_metadata(fmt):
            questionary.print(
                f"Removed non-valid metadata from '{comic.path.name}'.",
                style=Styles.WARNING,
            )

    def _sort_comic_list(self: Runner, file_list: list[Path]) -> None:
        """Sort comic archives in the provided list.

        This method sorts the comic archives in the file list based on the specified destination directory,
        using the configured sorting settings.

        Args:
            file_list: list[Path]: The list of comic archive file paths to sort.

        Returns:
            None
        """

        if not self.config.sort_dir:
            questionary.print(
                "\nUnable to sort files. No destination directory was provided.",
                style=Styles.ERROR,
            )
            return

        questionary.print(
            "\nStarting sorting of comic archives:\n----------------------------------",
            style=Styles.TITLE,
        )
        file_sorter = FileSorter(self.config.sort_dir)
        for comic in file_list:
            result = file_sorter.sort_comics(comic)
            if not result:
                questionary.print(f"Unable to move '{comic.name}'.", style=Styles.ERROR)

    def _comics_with_no_metadata(self: Runner, file_list: list[Path]) -> None:
        """Display files without metadata.

        This static method prints out the files in the provided list that do not have associated metadata.

        Args:
            file_list: list[Path]: The list of comic archive file paths to check for metadata.

        Returns:
            None
        """

        questionary.print(
            "\nShowing files without metadata:\n-------------------------------",
            style=Styles.TITLE,
        )

        if not (self.config.use_comic_info or self.config.use_metron_info):
            return

        for comic in file_list:
            comic_archive = Comic(str(comic))
            if (
                self.config.use_comic_info
                and not comic_archive.has_metadata(MetadataFormat.COMIC_RACK)
            ) or (
                self.config.use_metron_info
                and not comic_archive.has_metadata(MetadataFormat.METRON_INFO)
            ):
                questionary.print(f"{comic}", style=Styles.SUCCESS)

    def _delete_metadata(self: Runner, file_list: list[Path]) -> None:
        """Remove metadata from comic archives.

        This static method removes metadata from the comic archives in the provided list, if metadata exists.

        Args:
            file_list: list[Path]: The list of comic archive file paths to remove metadata from.

        Returns:
            None
        """

        questionary.print("\nRemoving metadata:\n-----------------", style=Styles.TITLE)
        for item in file_list:
            comic_archive = Comic(item)
            formats_removed = []

            if self.config.use_comic_info and comic_archive.has_metadata(
                MetadataFormat.COMIC_RACK
            ):
                comic_archive.remove_metadata(MetadataFormat.COMIC_RACK)
                formats_removed.append("'ComicInfo.xml'")

            if self.config.use_metron_info and comic_archive.has_metadata(
                MetadataFormat.METRON_INFO
            ):
                comic_archive.remove_metadata(MetadataFormat.METRON_INFO)
                formats_removed.append("'MetronInfo.xml'")

            if formats_removed:
                fmt = " and ".join(formats_removed)
                msg = f"Removed {fmt} metadata from '{item.name}'."
                questionary.print(msg, style=Styles.SUCCESS)
            else:
                questionary.print(f"no metadata in '{item.name}'", style=Styles.WARNING)

    def _has_credentials(self: Runner) -> bool:
        """Check if Metron credentials are present.

        This method returns True if Metron credentials are provided in the configuration settings, otherwise False.

        Returns:
            bool: True if Metron credentials are present, False otherwise.
        """

        return bool(self.config.metron_user and self.config.metron_pass)

    def _get_credentials(self: Runner) -> bool:
        """Prompt the user to enter Metron credentials.

        This method asks the user to input their Metron username and password, and provides an option to save the
        credentials based on the user's choice.

        Returns:
            bool: True if the user entered both username and password, False otherwise.
        """

        answers = questionary.form(
            user=questionary.text("What is your Metron username?"),
            passwd=questionary.text("What is your Metron password?"),
            save=questionary.confirm("Would you like to save your credentials?"),
        ).ask()
        if answers["user"] and answers["passwd"]:
            self.config.metron_user = answers["user"]
            self.config.metron_pass = answers["passwd"]
            if answers["save"]:
                self.config.save()
            return True
        return False

    def _get_sort_dir(self: Runner) -> bool:
        """Prompt the user to specify the directory for sorting comics.

        This method asks the user to input the directory path where the comics should be sorted to, and provides an
        option to save this location for future use.

        Returns:
            bool: True if a directory path is provided, False otherwise.
        """

        answers = questionary.form(
            dir=questionary.path("What directory should comics be sorted to?"),
            save=questionary.confirm("Would you like to save this location for future use?"),
        ).ask()
        if answers["dir"]:
            self.config.sort_dir = answers["dir"]
            if answers["save"]:
                self.config.save()
            return True
        return False

    @staticmethod
    def _get_duplicate_entry_index(
        comic_path: str,
        dups_list: list[DuplicateIssue],
    ) -> int | None:
        """Get the index of a duplicate entry in the list.

        This static method searches for the index of a duplicate entry in the list of DuplicateIssue objects based on
        the provided comic path.

        Args:
            comic_path: str: The path of the comic to search for.
            dups_list: list[DuplicateIssue]: The list of DuplicateIssue objects to search within.

        Returns:
            int | None: The index of the duplicate entry if found, otherwise None.
        """

        return next(
            (idx for idx, item in enumerate(dups_list) if comic_path in item.path_),
            None,
        )

    @staticmethod
    def _update_ci_xml(file_list: list[DuplicateIssue]) -> None:
        """Update ComicInfo metadata in comic archives.

        This static method updates the ComicInfo metadata in the provided list of comic archives with default page
        list information, if metadata exists.

        Args:
            file_list: list[DuplicateIssue]: The list of DuplicateIssue objects representing comic archives.

        Returns:
            None
        """

        for item in tqdm(file_list):
            comic = Comic(item.path_)
            if comic.has_metadata(MetadataFormat.COMIC_RACK):
                md = comic.read_metadata(MetadataFormat.COMIC_RACK)
                new_md = Metadata()
                new_md.set_default_page_list(comic.get_number_of_pages())
                md.overlay(new_md)
                if not comic.write_metadata(md, MetadataFormat.COMIC_RACK):
                    LOGGER.error("Could not write metadata to %s", comic)

    def _remove_duplicates(self: Runner, file_list: list[Path]) -> None:
        """Remove duplicate images from comic archives.

        This method identifies and allows the user to review and remove duplicate images from the provided list of
        comic archives, with options to delete pages per book and update the ComicInfo metadata.

        Args:
            file_list: list[Path]: The list of comic archive file paths to check for duplicates.

        Returns:
            None
        """

        dups_obj = Duplicates(file_list)
        distinct_hashes = dups_obj.get_distinct_hashes()
        if not questionary.confirm(
            f"Found {len(distinct_hashes)} duplicate images. Do you want to review them?",
        ).ask():
            return

        # List of page indexes to delete for each comic.
        duplicates_lst: list[DuplicateIssue] = []
        # This Loop runs for each *distinct* hash.
        for count, img_hash in enumerate(distinct_hashes, 1):
            comics_lst = dups_obj.get_comic_list_from_hash(img_hash)
            # Simple info message to show the user
            comics_lst_str = "\n\t".join(
                map(str, [Path(i.path_).name for i in comics_lst]),
            )
            questionary.print(
                f"Showing image #{count} which is in:\n\t{comics_lst_str}",
                style=Styles.SUCCESS,
            )
            # Get image from the first comic in the comics list for the hash to display
            # to the user.
            first_comic = dups_obj.get_comic_info_for_distinct_hash(img_hash)
            dups_obj.show_image(first_comic)

            # TODO: Give user the option to delete page per book.
            if questionary.confirm("Do you want to remove this image from all comics?").ask():
                for comic in comics_lst:
                    if duplicates_lst:
                        dup_entry_idx = self._get_duplicate_entry_index(
                            comic.path_,
                            duplicates_lst,
                        )
                        if dup_entry_idx is not None:
                            di: DuplicateIssue = duplicates_lst[dup_entry_idx]
                            di.pages_index.append(comic.pages_index[0])
                        else:
                            duplicates_lst.append(
                                DuplicateIssue(comic.path_, [comic.pages_index[0]]),
                            )
                    else:
                        duplicates_lst.append(
                            DuplicateIssue(comic.path_, [comic.pages_index[0]]),
                        )

        # After building the list let's ask the user if they want to write the changes.
        if (
            duplicates_lst
            and questionary.confirm("Do want to write your changes to the comics?").ask()
        ):
            dups_obj.delete_comic_pages(duplicates_lst)
        else:
            questionary.print("No duplicate page changes to write.", style=Styles.SUCCESS)

        # Ask user if they want to update ComicInfo.xml pages for changes. Not necessary for MetronInfo.xml
        if (
            self.config.use_comic_info
            and questionary.confirm(
                "Do you want to update the comic's 'comicinfo.xml' for the changes?",
            ).ask()
        ):
            self._update_ci_xml(duplicates_lst)

    def _no_md_fmt_set(self: Runner) -> bool:
        if not self.config.use_metron_info and not self.config.use_comic_info:
            questionary.print("No metadata format was given. Exiting...", style=Styles.ERROR)
            return True
        return False

    def run(self: Runner) -> None:  # noqa: PLR0912
        """Run Metron Tagger operations based on the provided settings.

        This method orchestrates various operations such as handling missing metadata, exporting to cbz,
        deleting metadata, removing duplicates, processing Ids or online, renaming comics, sorting, and validating
        comic information.

        Returns:
            None
        """

        if not (file_list := get_recursive_filelist(self.config.path)):
            questionary.print("No files to process. Exiting.", style=Styles.WARNING)
            sys.exit(0)

        # Start logging
        init_logging(self.config)

        if self.config.missing:
            if self._no_md_fmt_set():
                sys.exit(0)
            self._comics_with_no_metadata(file_list)

        if self.config.export_to_cbz:
            self._export_to_zip(file_list)

        if self.config.delete:
            if self._no_md_fmt_set():
                sys.exit(0)
            self._delete_metadata(file_list)

        if self.config.duplicates:
            self._remove_duplicates(file_list)

        if self.config.id or self.config.online:
            if not self._has_credentials() and not self._get_credentials():
                questionary.print("No credentials provided. Exiting...", style=Styles.ERROR)
                sys.exit(0)

            if self._no_md_fmt_set():
                sys.exit(0)

            t = Talker(
                self.config.metron_user,
                self.config.metron_pass,
                self.config.use_metron_info,
                self.config.use_comic_info,
            )
            if self.config.id:
                if len(file_list) == 1:
                    t.retrieve_single_issue(file_list[0], self.config.id)
                else:
                    questionary.print(
                        "More than one file was passed for Id processing. Exiting...",
                        style=Styles.WARNING,
                    )
                    sys.exit(0)
            else:
                t.identify_comics(file_list, self.config)

        if self.config.rename:
            file_list = self.rename_comics(file_list)

        if self.config.sort:
            if not self.config.sort_dir and not self._get_sort_dir():
                questionary.print("No sort directory given. Exiting...")
                sys.exit(0)
            self._sort_comic_list(file_list)

        if self.config.validate:
            if self._no_md_fmt_set():
                sys.exit(0)
            self._validate_comic_info(file_list, self.config.remove_non_valid)



from pathlib import Path

from metrontagger.settings import MetronTaggerSettings


def test_settings(tmpdir: Path) -> None:
    user = "test"
    dummy = "dummy_value"
    padding = 4
    cleanup = False
    file_template = "%series% v%volume% #%issue% (of %issuecount%) (%year%)"

    config = MetronTaggerSettings(config_dir=str(tmpdir))
    # Make sure initial values are correct
    assert not config.metron_user
    assert not config.metron_pass
    assert not config.sort_dir
    assert config.rename_issue_number_padding == 3  # noqa: PLR2004
    assert config.rename_use_smart_string_cleanup is True
    assert config.rename_template == "%series% v%volume% #%issue% (%year%)"

    # Save the new values
    config.metron_user = user
    config.metron_pass = dummy
    config.rename_issue_number_padding = padding
    config.rename_use_smart_string_cleanup = cleanup
    config.rename_template = file_template
    config.save()
    # Now load that file and verify the contents
    new_config = MetronTaggerSettings(config_dir=str(tmpdir))
    assert new_config.metron_user == user
    assert new_config.metron_pass == dummy
    assert not new_config.sort_dir
    assert new_config.rename_issue_number_padding == padding
    assert new_config.rename_use_smart_string_cleanup == cleanup
    assert new_config.rename_template == file_template



import sys
from datetime import date, datetime, timedelta, timezone
from decimal import Decimal
from pathlib import Path
from zipfile import ZipFile

import pytest
from darkseid.comic import Comic, MetadataFormat
from darkseid.metadata import AgeRatings, Basic, InfoSources, Links, Metadata, Notes, Series
from mokkari.schemas.base import BaseResource
from mokkari.schemas.generic import GenericItem
from mokkari.schemas.issue import BaseIssue, BasicSeries, Credit, Issue, IssueSeries
from mokkari.schemas.reprint import Reprint
from mokkari.session import Session
from pydantic import HttpUrl, TypeAdapter

from metrontagger.talker import InfoSource, Talker

tzinfo = timezone(timedelta(hours=-5))


@pytest.fixture()
def test_issue() -> Issue:
    issue = Issue(
        id=31047,
        publisher=GenericItem(id=1, name="Marvel"),
        series=IssueSeries(
            id=2222,
            name="The Spectacular Spider-Man",
            sort_name="Spectacular Spider-Man",
            volume=1,
            series_type=GenericItem(id=2, name="Single Issue"),
            year_began=1978,
        ),
        number="47",
        collection_title="",
        story_titles=["A Night on the Prowl!"],
        cover_date=date(1980, 10, 1),
        store_date=None,
        price=Decimal(".5"),
        rating=GenericItem(id=6, name="CCA"),
        upc="",
        isbn="",
        sku="",
        page_count=36,
        desc="Spider-Man goes on a wild goose chase to find out who is behind the Prowler impersonation.",
        image=HttpUrl(
            "https://static.metron.cloud/media/issue/2021/05/22/the-spectacular-spider-man-47.jpg"
        ),
        cover_hash="c0f83fe438876c1b",
        credits=[
            Credit(id=2335, creator="Al Milgrom", role=[GenericItem(id=7, name="Cover")]),
            Credit(
                id=1402,
                creator="Bruce Patterson",
                role=[GenericItem(id=4, name="Inker"), GenericItem(id=6, name="Letterer")],
            ),
            Credit(id=128, creator="Dennis O'Neil", role=[GenericItem(id=8, name="Editor")]),
            Credit(id=624, creator="Glynis Oliver", role=[GenericItem(id=5, name="Colorist")]),
            Credit(
                id=624,
                creator="Jim Shooter",
                role=[GenericItem(id=20, name="Editor In Chief")],
            ),
            Credit(
                id=675, creator="Marie Severin", role=[GenericItem(id=3, name="Penciller")]
            ),
            Credit(id=675, creator="Roger Stern", role=[GenericItem(id=1, name="Writer")]),
        ],
        characters=[
            BaseResource(id=6784, name="Debra Whitman", modified=datetime.now(tzinfo)),
            BaseResource(id=3067, name="Hobgoblin (Kingsley)", modified=datetime.now(tzinfo)),
            BaseResource(id=2415, name="Prowler", modified=datetime.now(tzinfo)),
            BaseResource(id=145, name="Spider-Man", modified=datetime.now(tzinfo)),
        ],
        cv_id=20745,
        resource_url=HttpUrl("https://metron.cloud/issue/the-spectacular-spider-man-1976-47/"),
        modified=datetime.now(tzinfo),
    )
    adapter = TypeAdapter(Issue)
    return adapter.validate_python(issue)


def create_reprint_list(resource: list[Reprint]) -> list[Basic]:
    return [Basic(r.issue, r.id) for r in resource]


def create_resource_list(resource: any) -> list[Basic]:
    return [Basic(r.name, r.id) for r in resource]


def create_read_md_resource_list(resource: any) -> list[Basic]:
    return [Basic(r.name) for r in resource]


def create_read_md_story(resource: list[str]) -> list[Basic]:
    return [Basic(story) for story in resource]


def test_map_resp_to_metadata(talker: Talker, test_issue: Issue) -> None:
    md = talker._map_resp_to_metadata(test_issue)
    assert md is not None
    assert md.stories == create_read_md_story(test_issue.story_titles)
    assert md.series.name == test_issue.series.name
    assert md.series.volume == test_issue.series.volume
    assert md.publisher.name == test_issue.publisher.name
    assert md.issue == test_issue.number
    assert md.story_arcs == create_resource_list(test_issue.arcs)
    assert md.teams == create_resource_list(test_issue.teams)
    assert md.cover_date.year == test_issue.cover_date.year
    assert md.characters == create_resource_list(test_issue.characters)
    assert md.credits is not None
    assert md.credits[0].person == "Al Milgrom"
    assert md.credits[0].role[0].name == "Cover"
    assert md.reprints == create_reprint_list(test_issue.reprints)
    assert md.age_rating == AgeRatings("Everyone", "Everyone")
    assert md.web_link == [
        Links("https://metron.cloud/issue/the-spectacular-spider-man-1976-47/", True)
    ]


def test_map_resp_to_metadata_with_no_story_name(
    talker: Talker,
    test_issue: Issue,
) -> None:
    test_data = test_issue
    test_data.story_titles = []
    meta_data = talker._map_resp_to_metadata(test_data)
    assert meta_data is not None
    assert len(meta_data.stories) == 0
    assert meta_data.series.name == test_issue.series.name
    assert meta_data.series.volume == test_issue.series.volume
    assert meta_data.series.start_year == test_issue.series.year_began
    assert meta_data.publisher.name == test_issue.publisher.name
    assert meta_data.issue == test_issue.number
    assert meta_data.cover_date.year == test_issue.cover_date.year


@pytest.fixture()
def test_issue_list() -> list[BaseIssue]:
    i_list = [
        BaseIssue(
            id=3634,
            series=BasicSeries(name="Aquaman", volume=1, year_began=1962),
            number="1",
            issue_name="Aquaman (1962) #1",
            cover_date=date(1962, 2, 1),
            image=HttpUrl(
                "https://static.metron.cloud/media/issue/2019/07/12/aquaman-v1-1.jpg"
            ),
            cover_hash="ccb2097c5b273c1b",
            modified=datetime.now(tzinfo),
        ),
        BaseIssue(
            id=2471,
            series=BasicSeries(name="Aquaman", volume=2, year_began=1986),
            number="1",
            issue_name="Aquaman (1986) #1",
            cover_date=date(1986, 2, 1),
            image=HttpUrl(
                "https://static.metron.cloud/media/issue/2019/05/19/aquaman-v2-1.jpg"
            ),
            cover_hash="ea97c11cb3660c79",
            modified=datetime.now(tzinfo),
        ),
        BaseIssue(
            id=2541,
            series=BasicSeries(name="Aquaman", volume=3, year_began=1989),
            number="1",
            issue_name="Aquaman (1989) #1",
            cover_date=date(1989, 6, 1),
            image=HttpUrl(
                "https://static.metron.cloud/media/issue/2019/05/25/aquaman-v3-1.jpg"
            ),
            cover_hash="d50df6181876276d",
            modified=datetime.now(tzinfo),
        ),
    ]
    adapter = TypeAdapter(list[BaseIssue])
    return adapter.validate_python(i_list)


def test_process_file(
    talker: Talker,
    fake_comic: ZipFile,
    test_issue_list: list[BaseIssue],
    mocker: any,
) -> None:
    # Remove any existing metadata from comic fixture
    ca = Comic(str(fake_comic))
    if ca.has_metadata(MetadataFormat.COMIC_RACK):
        ca.remove_metadata(MetadataFormat.COMIC_RACK)

    # Mock the call to Metron
    mocker.patch.object(Session, "issues_list", return_value=test_issue_list)
    talker._process_file(Path(str(fake_comic)), False)

    id_, multiple = talker._process_file(Path(str(fake_comic)), False)
    assert id_ is None
    assert multiple
    assert fake_comic in [c.filename for c in talker.match_results.multiple_matches]

    id_list = []
    for c in talker.match_results.multiple_matches:
        id_list.extend(i.id for i in c.matches)
    assert 2471 in id_list  # noqa: PLR2004


@pytest.mark.skipif(sys.platform in ["win32"], reason="Skip Windows.")
def test_write_issue_md(
    talker: Talker,
    fake_comic: ZipFile,
    test_issue: Issue,
    mocker: any,
) -> None:
    # Remove any existing metadata from comic fixture
    ca = Comic(str(fake_comic))
    if ca.has_metadata(MetadataFormat.COMIC_RACK):
        ca.remove_metadata(MetadataFormat.COMIC_RACK)

    # Mock the call to Metron
    mocker.patch.object(Session, "issue", return_value=test_issue)
    talker.retrieve_single_issue(Path(str(fake_comic)), 5)

    # Now let's test writing the metadata to file
    talker._write_issue_md(Path(str(fake_comic)), 1)
    ca = Comic(str(fake_comic))
    assert ca.has_metadata(MetadataFormat.COMIC_RACK)
    ca_md = ca.read_metadata(MetadataFormat.COMIC_RACK)
    assert ca_md.stories == create_read_md_story(test_issue.story_titles)
    assert ca_md.series.name == test_issue.series.name
    assert ca_md.series.volume == test_issue.series.volume
    assert ca_md.publisher.name == test_issue.publisher.name
    assert ca_md.issue == test_issue.number
    assert ca_md.teams is None
    assert ca_md.cover_date.year == test_issue.cover_date.year
    assert ca_md.characters == create_read_md_resource_list(test_issue.characters)
    assert ca_md.credits is not None
    assert ca_md.credits[0].person == "Roger Stern"
    assert ca_md.credits[0].role[0].name == "Writer"


@pytest.mark.skipif(sys.platform in ["win32"], reason="Skip Windows.")
def test_retrieve_single_issue(
    talker: Talker,
    fake_comic: ZipFile,
    test_issue: Issue,
    mocker: any,
) -> None:
    # Remove any existing metadata from comic fixture
    ca = Comic(str(fake_comic))
    if ca.has_metadata(MetadataFormat.COMIC_RACK):
        ca.remove_metadata(MetadataFormat.COMIC_RACK)

    # Mock the call to Metron
    mocker.patch.object(Session, "issue", return_value=test_issue)
    talker.retrieve_single_issue(Path(str(fake_comic)), 10)

    # Now let's test the metadata
    ca = Comic(str(fake_comic))
    assert ca.has_metadata(MetadataFormat.COMIC_RACK)
    ca_md = ca.read_metadata(MetadataFormat.COMIC_RACK)
    assert ca_md.stories == create_read_md_story(test_issue.story_titles)
    assert ca_md.series.name == test_issue.series.name
    assert ca_md.series.volume == test_issue.series.volume
    assert ca_md.publisher.name == test_issue.publisher.name
    assert ca_md.issue == test_issue.number
    assert ca_md.teams is None
    assert ca_md.cover_date.year == test_issue.cover_date.year
    assert ca_md.characters == create_read_md_resource_list(test_issue.characters)
    assert ca_md.credits is not None
    assert ca_md.credits[0].person == "Roger Stern"
    assert ca_md.credits[0].role[0].name == "Writer"


@pytest.mark.parametrize(
    ("primary_name", "primary_id", "alternatives", "expected"),
    [
        # Happy path tests
        ("metron", 123, [], (InfoSource.metron, 123)),
        ("comic vine", 456, [], (InfoSource.comic_vine, 456)),
        ("anilist", 789, [InfoSources("metron", 321)], (InfoSource.metron, 321)),
        # Edge case
        ("anilist", 789, [], None),
    ],
    ids=[
        "happy_path_metron_primary",
        "happy_path_comic_vine_primary",
        "happy_path_metron_alternative",
        "edge_case_no_metron_or_comic_vine",
    ],
)
def test_get_id_from_metron_info(primary_name, primary_id, alternatives, expected):
    # Arrange
    md = Metadata(info_source=[InfoSources(primary_name, primary_id, True), *alternatives])

    # Act
    result = Talker._get_id_from_metron_info(md)

    # Assert
    assert result == expected


@pytest.mark.parametrize(
    ("notes", "expected"),
    [
        # Happy path tests
        (
            "Tagged with MetronTagger-2.6.0 using info from Metron on 2024-10-16 16:07:08. [issue_id:12345]",
            (InfoSource.metron, 12345),
        ),
        (
            "Tagged with ComicTagger 1.3.2a5 using info from Comic Vine on 2022-04-16 15:52:26. [Issue ID 67890]",
            (InfoSource.comic_vine, 67890),
        ),
        # Edge cases
        (
            "Tagged with MetronTagger-2.6.0 using info from Metron on 2024-10-16 16:07:08. [issue_id:00001]",
            (InfoSource.metron, 1),
        ),
        (
            "Tagged with ComicTagger 1.3.2a5 using info from Comic Vine on 2022-04-16 15:52:26. [Issue ID 00000]",
            (InfoSource.comic_vine, 0),
        ),
        # Error cases
        (
            "Tagged with MetronTagger-2.6.0 using info from Metron on 2024-10-16 16:07:08. [issue_id:abcde]",
            None,
        ),
        (
            "Tagged with ComicTagger 1.3.2a5 using info from Comic Vine on 2022-04-16 15:52:26. [Issue ID abcde]",
            None,
        ),
        ("[unknown] Issue ID: 12345", None),
        (None, None),
    ],
    ids=[
        "metrontagger_valid_id",
        "comictagger_valid_id",
        "metrontagger_leading_zeros",
        "comictagger_zero_id",
        "metrontagger_invalid_id",
        "comictagger_invalid_id",
        "unknown_source",
        "no_notes",
    ],
)
def test_get_id_from_comic_info(notes, expected):
    # Arrange
    md = Metadata(
        notes=Notes(comic_rack=notes) if notes else None, series=Series("Foo"), issue="2"
    )

    # Act
    result = Talker._get_id_from_comic_info(md)

    # Assert
    assert result == expected



"""Logging module"""

import logging
from logging import basicConfig

from metrontagger.settings import MetronTaggerSettings

DATE_FMT = "%Y-%m-%d %H:%M:%S %Z"
LOG_FMT = "{asctime} {levelname:8} {message}"


def init_logging(config: MetronTaggerSettings) -> None:
    """Initializing logging"""
    formatter = logging.Formatter(LOG_FMT, style="{", datefmt=DATE_FMT)
    log_path = config.get_settings_folder() / "metron-tagger.log"
    handler = logging.FileHandler(str(log_path))
    handler.setFormatter(formatter)
    basicConfig(level=logging.WARNING, handlers=[handler])



from __future__ import annotations

import io
import warnings
from datetime import datetime
from enum import Enum, auto, unique
from logging import getLogger
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pathlib import Path

    from mokkari.schemas.generic import GenericItem
    from mokkari.schemas.issue import BaseIssue, Credit as MokkariCredit, Issue

    from metrontagger.settings import MetronTaggerSettings

import mokkari
import questionary
from comicfn2dict import comicfn2dict
from darkseid.comic import Comic, MetadataFormat
from darkseid.issue_string import IssueString
from darkseid.metadata import (
    AgeRatings,
    Arc,
    Basic,
    Credit,
    InfoSources,
    Links,
    Metadata,
    Notes,
    Publisher,
    Role,
    Series,
)
from imagehash import ImageHash, hex_to_hash, phash
from mokkari.exceptions import ApiError
from PIL import Image

from metrontagger import __version__
from metrontagger.styles import Styles
from metrontagger.utils import create_query_params

LOGGER = getLogger(__name__)

warnings.filterwarnings(
    "ignore", category=UserWarning
)  # Ignore 'UserWarning: Corrupt EXIF data' warnings

HAMMING_DISTANCE = 10


@unique
class InfoSource(Enum):
    """An enumeration of information sources.

    This class defines different sources of information, including Metron, Comic Vine, and Unknown.
    """

    metron = auto()
    comic_vine = auto()
    unknown = auto()


class MultipleMatch:
    """Class to store multiple matches for a filename.

    This class initializes with a filename and a list of BaseIssue matches.
    """

    def __init__(self: MultipleMatch, filename: Path, match_list: list[BaseIssue]) -> None:
        self.filename = filename
        self.matches = match_list


class OnlineMatchResults:
    """Class to store online match results.

    This class stores good matches, no matches, and multiple matches for online search results.
    """

    def __init__(self: OnlineMatchResults) -> None:
        """Initialize the OnlineMatchResults class.

        This method initializes lists to store good matches, no matches, and multiple matches.
        """
        self.good_matches: list[Path] = []
        self.no_matches: list[Path] = []
        self.multiple_matches: list[MultipleMatch] = []

    def add_good_match(self: OnlineMatchResults, file_name: Path) -> None:
        """Add a good match to the list.

        This method appends a file name to the list of good matches.
        """
        self.good_matches.append(file_name)

    def add_no_match(self: OnlineMatchResults, file_name: Path) -> None:
        """Add a no match to the list.

        This method appends a file name to the list of no matches.
        """
        self.no_matches.append(file_name)

    def add_multiple_match(self: OnlineMatchResults, multi_match: MultipleMatch) -> None:
        """Add a multiple match to the list.

        This method appends a MultipleMatch object to the list of multiple matches.
        """
        self.multiple_matches.append(multi_match)


class Talker:
    """Class for handling online search and tagging of comic files.

    This class provides methods for identifying comics, retrieving single issues, and processing match results.
    """

    def __init__(
        self: Talker, username: str, password: str, metron_info: bool, comic_info: bool
    ) -> None:
        """Initialize the Talker class with API credentials.

        This method sets up the API connection using the provided username and password, and initializes match
        results storage.
        """
        self.api = mokkari.api(username, password, user_agent=f"Metron-Tagger/{__version__}")
        self.metron_info = metron_info
        self.comic_info = comic_info
        self.match_results = OnlineMatchResults()

    @staticmethod
    def _create_choice_list(match_set: list[BaseIssue]) -> list[questionary.Choice]:
        """Create a list of choices for selecting issues.

        This static method generates a list of choices based on the provided list of BaseIssue objects, including
        issue names and cover dates.

        Args:
            match_set: list[BaseIssue]: The list of BaseIssue objects to create choices from.

        Returns:
            list[questionary.Choice]: A list of choices for selecting issues, including a 'Skip' option.
        """
        issue_lst = []
        for i in match_set:
            c = questionary.Choice(title=f"{i.issue_name} ({i.cover_date})", value=i.id)
            issue_lst.append(c)

        issue_lst.append(questionary.Choice(title="Skip", value=""))
        return issue_lst

    def _select_choice_from_matches(
        self: Talker,
        fn: Path,
        match_set: list[BaseIssue],
    ) -> int | None:
        """Select an issue from a list of matches.

        This method displays the results found for a file, sorts the match list by cover date, and prompts the user
        to select an issue to match.

        Args:
            fn: Path: The file path for which matches are being selected.
            match_set: list[BaseIssue]: The list of BaseIssue objects to choose from.

        Returns:
            int | None: The selected issue ID or None if no selection is made.
        """
        questionary.print(f"\n{fn.name} - Results found:", style=Styles.TITLE)

        # sort match list by cover date
        match_set = sorted(match_set, key=lambda m: m.cover_date)
        choices = self._create_choice_list(match_set)

        return questionary.select("Select an issue to match", choices=choices).ask()

    @staticmethod
    def _get_comic_cover_hash(comic: Comic) -> ImageHash | None:
        """Get the image hash of a comic cover.

        This static method calculates the image hash of the comic cover image using pHash algorithm.

        Args:
            comic: Comic: The Comic object representing the comic.

        Returns:
            ImageHash | None: The image hash of the comic cover, or None if unable to calculate.
        """
        with Image.open(io.BytesIO(comic.get_page(0))) as img:
            try:
                ch = phash(img)
            except OSError:
                questionary.print("Unable to get cover hash.", style=Styles.ERROR)
                ch = None
        return ch

    def _within_hamming_distance(
        self: Talker, comic: Comic, metron_hash: str | None = None
    ) -> bool:
        """Check if the comic cover hash is within the specified Hamming distance.

        This method compares the comic cover hash with a provided hash to determine if they are within the specified
        Hamming distance.
        """
        if metron_hash is None:
            return False
        comic_hash = self._get_comic_cover_hash(comic)
        if comic_hash is None:
            return False
        hamming = comic_hash - hex_to_hash(metron_hash)
        return hamming <= HAMMING_DISTANCE

    def _get_hamming_results(self: Talker, comic: Comic, lst: list[BaseIssue]) -> list[any]:
        """Get the list of BaseIssue objects within the specified Hamming distance.

        This method filters the list of BaseIssue objects based on the Hamming distance between the comic cover hash
        and the cover hash of each item.
        """
        return [item for item in lst if self._within_hamming_distance(comic, item.cover_hash)]

    @staticmethod
    def _get_id_from_metron_info(md: Metadata) -> None | tuple[InfoSource, int]:
        online_sources = {"metron", "comic vine"}
        for src in md.info_source:
            lower_name = src.name.lower()
            if lower_name in online_sources:
                return (
                    InfoSource.metron if lower_name == "metron" else InfoSource.comic_vine,
                    src.id_,
                )
        return None

    @staticmethod
    def _get_id_from_comic_info(md: Metadata) -> None | tuple[InfoSource, int]:
        def _extract_id_str(notes: str, keyword: str) -> str:
            return notes.split(keyword)[1].split("]")[0].strip()

        # If `Notes` element doesn't exist let's bail.
        if md.notes is None:
            return None

        lower_notes = md.notes.comic_rack.lower()

        if "metrontagger" in lower_notes:
            source = InfoSource.metron
            id_str = _extract_id_str(md.notes.comic_rack, "issue_id:")
        elif "comictagger" in lower_notes:
            if "metron" in lower_notes:
                source = InfoSource.metron
            elif "comic vine" in lower_notes:
                source = InfoSource.comic_vine
            else:
                source = InfoSource.unknown
            id_str = _extract_id_str(md.notes.comic_rack, "Issue ID")
        else:
            return None

        try:
            id_ = int(id_str)
        except ValueError:
            LOGGER.exception("Comic has invalid id: %s #%s", md.series.name, md.issue)
            return None

        return source, id_

    def _process_file(self: Talker, fn: Path, interactive: bool) -> tuple[int | None, bool]:  # noqa: PLR0912 PLR0911
        """Process a comic file for metadata.

        This method processes a comic file to extract metadata, including checking for existing metadata, extracting
        source and ID information, and handling multiple matches.

        Args:
            fn: Path: The file path of the comic to process.
            interactive: bool: A flag indicating if the process should be interactive.

        Returns: tuple[int | None, bool]: A tuple containing the issue ID and a flag indicating if multiple matches
        were found.
        """
        ca = Comic(fn)

        if not ca.is_writable() and not ca.seems_to_be_a_comic_archive():
            questionary.print(
                f"{fn.name} appears not to be a comic or writable.",
                style=Styles.ERROR,
            )
            return None, False

        # Check if comic has a MetronInfo.xml or ComicInfo.xml that contains either a cvid or metron id.
        source = id_ = None
        # Let's prefer MetronInfo.xml over ComicInfo.xml, since it's easier to get info.
        if ca.has_metadata(MetadataFormat.METRON_INFO):
            md = ca.read_metadata(MetadataFormat.METRON_INFO)
            if res := self._get_id_from_metron_info(md):
                source, id_ = res
        elif ca.has_metadata(MetadataFormat.COMIC_RACK):
            md = ca.read_metadata(MetadataFormat.COMIC_RACK)
            if res := self._get_id_from_comic_info(md):
                source, id_ = res

        if source is not None and id_ is not None:

            def _print_metadata_message(src: InfoSource, comic: Comic) -> None:
                source_messages = {
                    InfoSource.metron: "Metron ID",
                    InfoSource.comic_vine: "ComicVine",
                    InfoSource.unknown: "Unknown Source",
                }
                if src in source_messages:
                    if src != InfoSource.unknown:
                        questionary.print(
                            f"Found {source_messages[src]} in '{comic}' metadata and using that to get the metadata...",
                            style=Styles.INFO,
                        )
                    else:
                        questionary.print(
                            f"Found {source_messages[src]} in '{comic}' metadata. Skipping...",
                            style=Styles.WARNING,
                        )

            _print_metadata_message(source, ca)

            match source:
                case InfoSource.metron:
                    self.match_results.add_good_match(fn)
                    return id_, False
                case InfoSource.comic_vine:
                    issues = self.api.issues_list(params={"cv_id": id_})
                    # This should always be 1 otherwise let's do a regular search.
                    if len(issues) == 1:
                        return issues[0].id, False
                case _:
                    pass

        # Alright, if the comic doesn't have an let's do a search based on the filename.
        # TODO: Determine if we want to use some of the other keys beyond 'series' and 'issue number'
        metadata: dict[str, str | tuple[str, ...]] = comicfn2dict(fn, verbose=0)

        params = create_query_params(metadata)
        i_list = self.api.issues_list(params=params)
        result_count = len(i_list)

        # No matches
        if result_count <= 0:
            self.match_results.add_no_match(fn)
            return None, False

        # Multiple matches
        if result_count > 1:
            LOGGER.debug("Check Hamming for '%s'", ca)
            hamming_lst = self._get_hamming_results(ca, i_list)
            # Matched single cover within hamming distance from multiple results
            if hamming_lst and len(hamming_lst) == 1:
                self.match_results.add_good_match(fn)
                return hamming_lst[0].id, False
            # No hamming match, let's ask the user later.
            self.match_results.add_multiple_match(MultipleMatch(fn, i_list))
            return None, True

        # Ask for user interaction for each comic.
        if interactive:
            issue_id = self._select_choice_from_matches(fn, i_list)
            if issue_id:
                self.match_results.add_good_match(fn)
                return issue_id, False
            self.match_results.add_multiple_match(MultipleMatch(fn, i_list))
            return None, True

        # Single match within the hamming distance
        if self._within_hamming_distance(ca, i_list[0].cover_hash):
            self.match_results.add_good_match(fn)
            return i_list[0].id, False

        # Single match not withing the hamming distance. We'll ask if the single choice is correct.
        self.match_results.add_multiple_match(MultipleMatch(fn, i_list))
        return None, True

    def _post_process_matches(self: Talker) -> None:
        """Post-process the match results.

        This method prints the successful matches and no matches, and handles files with multiple matches by
        selecting an issue to write metadata for.
        """
        # Print file matching results.
        if self.match_results.good_matches:
            questionary.print("\nSuccessful matches:\n------------------", style=Styles.TITLE)
            for comic in self.match_results.good_matches:
                questionary.print(f"{comic}", style=Styles.SUCCESS)

        if self.match_results.no_matches:
            questionary.print("\nNo matches:\n------------------", style=Styles.TITLE)
            for comic in self.match_results.no_matches:
                questionary.print(f"{comic}", style=Styles.WARNING)

        # Handle files with multiple matches
        if self.match_results.multiple_matches:
            for match_set in self.match_results.multiple_matches:
                if issue_id := self._select_choice_from_matches(
                    match_set.filename,
                    match_set.matches,
                ):
                    self._write_issue_md(match_set.filename, issue_id)

    def _write_issue_md(self, filename: Path, issue_id: int) -> None:
        """Write metadata for a comic issue.

        This method retrieves issue data from an API, overlays it with existing metadata, and writes the metadata to
        the specified comic file. It handles potential errors during data retrieval and provides feedback on the
        success or failure of the metadata writing process.

        Args:
            filename (Path): The path to the comic file where metadata will be written.
            issue_id (int): The identifier for the comic issue to retrieve metadata for.

        Returns:
            None

        Raises:
            ApiError: If there is an error retrieving data from the API.

        Examples:
            # Example usage:
            # talker_instance._write_issue_md(Path("comic.cbz"), 123)
        """

        try:
            resp = self.api.issue(issue_id)
        except ApiError as e:
            questionary.print(f"Failed to retrieve data: {e!r}", style=Styles.ERROR)
            return

        ca = Comic(filename)
        meta_data = Metadata()
        meta_data.set_default_page_list(ca.get_number_of_pages())
        md = self._map_resp_to_metadata(resp)
        md.overlay(meta_data)

        md_fmt = []
        if self.comic_info and ca.write_metadata(md, MetadataFormat.COMIC_RACK):
            md_fmt.append("'ComicInfo.xml'")
        if self.metron_info and ca.write_metadata(md, MetadataFormat.METRON_INFO):
            md_fmt.append("'MetronInfo.xml'")

        if md_fmt:
            collection = md.series.format.lower() in ["trade paperback", "hard cover"]
            collection_text = " (Collection)" if collection else ""
            fmt = " and ".join(md_fmt)
            msg = (
                f"Writing {fmt} metadata using "
                f"'{md.series.name} #{md.issue} ({md.cover_date.year}){collection_text}' "
                f"for '{filename.name}'."
            )
            questionary.print(msg, style=Styles.SUCCESS)
        else:
            questionary.print(
                f"There was a problem writing metadata for '{filename.name}'.",
                style=Styles.ERROR,
            )

    def identify_comics(
        self: Talker,
        file_list: list[Path],
        config: MetronTaggerSettings,
    ) -> None:
        """Identify and tag comics from a list of files.

        This method initiates an online search and tagging process for each file in the provided list, skipping files
        with existing metadata and handling multiple matches.

        Args:
            file_list: list[Path]: A list of file paths to process.
            config: MetronTaggerSettings: The configuration settings for the tagging process.

        Returns:
            None
        """
        questionary.print(
            "\nStarting online search and tagging:\n----------------------------------",
            style=Styles.TITLE,
        )

        for fn in file_list:
            if config.ignore_existing:
                comic = Comic(fn)
                if (
                    comic.has_metadata(MetadataFormat.COMIC_RACK)
                    and self.comic_info
                    and not self.metron_info
                ) or (
                    comic.has_metadata(MetadataFormat.METRON_INFO)
                    and self.metron_info
                    and not self.comic_info
                ):
                    questionary.print(
                        f"{fn.name} has metadata. Skipping...",
                        style=Styles.WARNING,
                    )
                    continue

            issue_id, multiple_match = self._process_file(fn, config.interactive)
            if issue_id:
                self._write_issue_md(fn, issue_id)
            elif not multiple_match:
                questionary.print(f"No Match for '{fn.name}'.", style=Styles.ERROR)

        # Print match results
        self._post_process_matches()

    def retrieve_single_issue(self: Talker, fn: Path, id_: int) -> None:
        """Retrieve and write metadata for a single issue.

        This method retrieves metadata for a single issue using the provided ID and writes the metadata to the
        corresponding file.
        """
        self._write_issue_md(fn, id_)

    @staticmethod
    def _map_resp_to_metadata(resp: Issue) -> Metadata:  # noqa: PLR0915
        """Map response data to metadata.

        This static method maps the response data from an Issue object to a Metadata object, including information
        such as series, issue number, publisher, cover date, and additional metadata details.
        """

        # Helper functions
        def create_resource_list(resource: any) -> list[Basic]:
            """Create a list of Basic objects from a given resource.

            This function takes a resource and generates a list of Basic objects with name and ID attributes.
            """
            return [Basic(r.name, r.id) for r in resource]

        def create_arc_list(resource: any) -> list[Arc]:
            """Create a list of Arc objects from a given resource.

            This function takes a resource containing items and constructs a list of Arc
            objects using the attributes of each item. It ensures that the 'number' attribute
            is set to None if it is not provided.

            Args:
                resource (any): An iterable containing items with 'name', 'id', and 'number' attributes.

            Returns:
                list[Arc]: A list of Arc objects created from the provided resource.
            """
            return [Arc(item.name, item.id) for item in resource]

        def create_notes(issue_id: int) -> Notes:
            """Create a note for an issue.

            This function generates a note string including tagging information from MetronTagger and the issue ID.
            """
            now_date = datetime.now().strftime("%Y-%m-%d %H:%M:%S")  # noqa: DTZ005
            metron_info_note = (
                f"Tagged with MetronTagger-{__version__} using info from Metron on {now_date}."
            )
            comic_rack_note = f"{metron_info_note} [issue_id:{issue_id}]"
            return Notes(metron_info=metron_info_note, comic_rack=comic_rack_note)

        def add_credits_to_metadata(
            meta_data: Metadata,
            credits_resp: list[MokkariCredit],
        ) -> Metadata:
            """Add credits to metadata.

            This function adds credits information to the metadata object based on the provided credits response.

            Args:
                meta_data: Metadata: The metadata object to add credits to.
                credits_resp: list[MokkariCredit]: The list of credits information to add.

            Returns:
                Metadata: The updated metadata object with added credits.
            """

            def create_role_list(roles: list[GenericItem]) -> list[Role]:
                """Create a list of Role objects from a list of GenericItem objects.

                This function converts a list of GenericItem objects representing roles into a list of Role objects
                with name and ID attributes.
                """
                return [Role(r.name, r.id) for r in roles]

            for c in credits_resp:
                if c.role:
                    meta_data.add_credit(Credit(c.creator, create_role_list(c.role), c.id))
                else:
                    meta_data.add_credit(Credit(c.creator, [], c.id))
            return meta_data

        def map_ratings(rating: str) -> AgeRatings:
            """Map a rating string to a standardized format.

            This function maps a given rating string to a standardized format ('Everyone', 'Teen', 'Mature 17+',
            or 'Unknown').

            Args:
                rating: str: The rating string to map.

            Returns:
                str: The standardized rating.
            """
            age_rating = rating.lower()
            if age_rating in {"everyone", "cca"}:
                return AgeRatings(metron_info="Everyone", comic_rack="Everyone")
            if age_rating in {"teen", "teen plus"}:
                return AgeRatings(metron_info="Teen", comic_rack="Teen")
            if age_rating in {"mature"}:
                return AgeRatings(metron_info="Mature", comic_rack="Mature 17+")
            return AgeRatings(metron_info="Unknown", comic_rack="Unknown")

        md = Metadata()

        alt_info_source = [InfoSources("Comic Vine", resp.cv_id)] if resp.cv_id else []
        md.info_source = [InfoSources("Metron", resp.id, True)] + alt_info_source  # NOQA: RUF005
        md.series = Series(
            name=resp.series.name,
            id_=resp.series.id,
            sort_name=resp.series.sort_name,
            volume=resp.series.volume,
            format=resp.series.series_type.name,
            start_year=resp.series.year_began,
        )
        md.issue = IssueString(resp.number).as_string() if resp.number else None
        md.publisher = (
            Publisher(resp.publisher.name, resp.publisher.id) if resp.publisher else None
        )
        md.publisher.imprint = (
            Basic(resp.imprint.name, resp.imprint.id) if resp.imprint else None
        )
        md.cover_date = resp.cover_date or None
        md.store_date = resp.store_date or None
        md.comments = resp.desc
        md.notes = create_notes(resp.id)
        md.modified = resp.modified
        if resp.story_titles:
            md.stories = [Basic(story) for story in resp.story_titles]
        if resp.characters:
            md.characters = create_resource_list(resp.characters)
        if resp.teams:
            md.teams = create_resource_list(resp.teams)
        if resp.arcs:
            md.story_arcs = create_arc_list(resp.arcs)
        if resp.series.genres:
            md.genres = create_resource_list(resp.series.genres)
        if resp.reprints:
            md.reprints = [Basic(r.issue, r.id) for r in resp.reprints]
        if resp.credits:
            md = add_credits_to_metadata(md, resp.credits)
        if resp.rating:
            md.age_rating = map_ratings(resp.rating.name)
        if resp.resource_url:
            md.web_link = [Links(str(resp.resource_url), True)]

        return md
